import argparse
import glob
import os
import sys
from typing import List

from valohai.config import is_running_in_valohai
from valohai.internals import global_state
from valohai.internals.input_info import FileInfo, InputInfo
from valohai.parameters import Parameter


def prepare(*, step: str, default_parameters: dict = {}, default_inputs: dict = {}, image: str = None):
    """Define the name of the step and it's required inputs, parameters and Docker image

    Has dual purpose:
    - Provide default values for inputs, parameters and Docker image so the user code can be executed
    - Provide entry-point for the parser that generates/updates valohai.yaml integration file

    :param step: Step name for valohai.yaml
    :param default_parameters: Dict of parameters and default values
    :param default_inputs: Dict of inputs with (list of) default URIs
    :param image: Default docker image

    """
    global_state.step_name = step
    global_state.image_name = image

    parser = argparse.ArgumentParser()
    for name, default_value in dict(default_inputs).items():
        parser.add_argument(f"--{name}", type=str, nargs="+", default=default_value)
    for name, default_value in dict(default_parameters).items():
        parser.add_argument(
            f"--{name}", type=type(default_value), default=default_value
        )
    known_args, unknown_args = parser.parse_known_args()

    if not is_running_in_valohai():
        _load_inputs(known_args, list(default_inputs.keys()))
    _load_parameters(known_args, list(default_parameters.keys()))

    for unknown in unknown_args:
        print(
            f"Warning: Unexpected command-line argument {unknown} found.",
            file=sys.stderr,
        )


def _load_inputs(args: argparse.Namespace, names: List[str]):
    """Pull inputs from the command-line args

    User provides inputs and their default values in a dict, when calling prepare().
    Here we parse possible overrides from the command-line args.

    This is only ran for local executions. Cloud execution will get this same info directly from inputs.json.

    :param names: List of all possible input names

    """
    for name, values in vars(args).items():
        if name not in names:
            continue

        if not isinstance(values, list):
            values = [values]

        files = []
        for value in values:
            if "://" not in value:  # The string is a local path
                for path in glob.glob(value):
                    files.append(
                        FileInfo(
                            name=os.path.basename(path),
                            uri=None,
                            path=value,
                            size=None,
                            checksums=None,
                        )
                    )
            else:  # The string is an URI
                files.append(
                    FileInfo(
                        name=uri_to_filename(value),
                        uri=value,
                        path=None,
                        size=None,
                        checksums=None,
                    )
                )

        input_info = InputInfo(files)
        global_state.input_infos[name] = input_info


def _load_parameters(args: argparse.Namespace, names: List[str]):
    """Pull parameters from the command-line args

    User provides parameters and their default values in a dict, when calling prepare().
    Here we parse possible overrides from the command-line args.

    :param args: All command-line arguments
    :param names: List of all possible parameters names

    """
    for name, value in vars(args).items():
        if name not in names:
            continue
        Parameter(name).value = value


def uri_to_filename(uri: str) -> str:
    return uri.rpartition("/")[-1]



import json

_supported_types = [int, float]


class Logger:
    def __init__(self):
        self.partial_logs = {}

    def __enter__(self):
        self.partial_logs = {}
        return self

    def __exit__(self, type, value, traceback):
        self.flush()

    def log(self, *args, **kwargs):
        """Log a single name/value pair to be flushed into standard output later as batch.

        For a repeating iteration like a machine learning training loop, Valohai expects
        all logged values to be printed as a batch.

        Example:
            for epoch in range(10):
                with valohai.logger() as logger:
                    logger.log("epoch", epoch)
                    logger.log("accuracy", 0.54)
                    logger.log("loss", 0.123)

        Example 2:
            for epoch in range(10):
                logger = valohai.logger():
                logger.log("epoch", epoch)
                logger.log("accuracy", 0.54)
                logger.log("loss", 0.123)
                logger.flush_logs()

        Example 3:
            for epoch in range(10):
                with valohai.logger() as logger:
                    logger.log("epoch", epoch)
                    logger.log(acc=0.54, loss=0.123)

        All three examples will act exactly the same.

        :param name: Name of the variable being logged (example: learning_rate)
        :param value: Value of the logged variable

        """
        if len(args) % 2 != 0:
            raise ValueError(f"Odd number of arguments in {args} for log()")
        for key, value in zip(args[::2], args[1::2]):
            self._serialize(key, value)
        for key, value in kwargs.items():
            self._serialize(key, value)

    def flush(self):
        """Flush all the partial logs into standard as a batch.

        For a repeating iteration like a machine learning training loop, Valohai expects
        all logged values to be printed as single batch.

        Example:
            logger = valohai.logger():
            logger.log("epoch", epoch)
            logger.log("accuracy", 0.54)
            logger.log("loss", 0.123)
            logger.flush_logs()

        This will log all three metrics at once.
        """
        if self.partial_logs:
            # Start with \n, ensuring JSON prints on it's own line
            print(f'\n{json.dumps(self.partial_logs, default=str)}')
            self.partial_logs.clear()

    def _serialize(self, name, value):
        self.partial_logs.update({str(name): value})


logger = Logger



import copy

from valohai_yaml.objs import Config, Step
from valohai_yaml.objs.base import Item
from valohai_yaml.utils.merge import merge_dicts, merge_simple

from valohai.consts import DEFAULT_DOCKER_IMAGE

def python_to_yaml_merge_strategy(original: Item, parsed: Item) -> Item:
    """Merging strategy in the valohai-utils AST parser use-case

    :param original: Original Item from an existing valohai.yaml
    :param parsed: New Item parsed from a .py file using valohai-utils AST parser
    :return: Merged Item
    """
    if isinstance(parsed, Config) and isinstance(original, Config):
        return _merge_config(original, parsed)
    if isinstance(parsed, Step) and isinstance(original, Step):
        return _merge_step(original, parsed)
    return original.merge_with(parsed)


def _merge_config(original: Config, parsed: Config) -> Config:
    """Merging strategy for Configs in the valohai-utils AST parser use-case

    :param original: Original Config from an existing valohai.yaml
    :param parsed: New Config parsed from a .py file using valohai-utils AST parser
    :return: Merged Config
    """
    result = copy.deepcopy(original)
    for key, step in parsed.steps.items():
        if key in result.steps:
            result.steps[key] = original.steps[key].merge_with(
                step, python_to_yaml_merge_strategy
            )
        else:
            result.steps[key] = step
    return result


def _merge_step(original: Step, parsed: Step) -> Step:
    """Merging strategy for Steps in the valohai-utils AST parser use-case

    :param original: Original Step from an existing valohai.yaml
    :param parsed: New Step parsed from a .py file using valohai-utils AST parser
    :return: Merged Step
    """

    # AST parser overrides parameters, inputs and step name. Respect the original values for everything else.
    result = Step(
        name=parsed.name,
        image=original.image,
        command=original.command,
        environment=original.environment,
        description=original.description,
        outputs=original.outputs,
        mounts=original.mounts,
    )

    # If user first types "learning_rage", creates config, and finally fixes the typo,
    # they don't want to end up with both "learning_rage" and "learning_rate" after the merge.
    # So only merge parameters and inputs that are part of the new config (skip_missing_b=True)
    result.parameters = merge_dicts(
        a=original.parameters,
        b=parsed.parameters,
        merger=merge_simple,
        copier=copy.deepcopy,
        skip_missing_b=True,
    )
    result.inputs = merge_dicts(
        a=original.inputs,
        b=parsed.inputs,
        merger=merge_simple,
        copier=copy.deepcopy,
        skip_missing_b=True,
    )

    if parsed.image is not None and parsed.image != DEFAULT_DOCKER_IMAGE :
        result.image = parsed.image

    return result



import json
import os
from typing import Union

from valohai.internals import global_state
from valohai.paths import get_parameters_config_path

supported_types = Union[int, float, bool, str, None]


def load_parameter(name: str, default: supported_types) -> supported_types:
    parameters_config_path = get_parameters_config_path()
    if os.path.isfile(parameters_config_path):
        with open(parameters_config_path) as json_file:
            data = json.load(json_file)
            if name in data:
                global_state.parsed_parameters[name] = data[name]
                return global_state.parsed_parameters[name]
    return default



import ast
import os
import re

import setuptools

with open(os.path.join(os.path.dirname(__file__), 'valohai', '__init__.py')) as infp:
    version = ast.literal_eval(
        re.search('__version__ = (.+?)$', infp.read(), re.M).group(1)
    )

setuptools.setup(
    name="valohai-utils",
    version=version,
    author="Valohai",
    author_email="hait@valohai.com",
    license="MIT",
    packages=setuptools.find_packages(include=("valohai*",)),
    install_requires=[
        "tqdm",
        "requests",
        "valohai-yaml>=0.13.0",
    ],
    python_requires='>=3.6',
)
